/*============================================================================
  File:     DropTableAndRecoverToPointInTime-MoreComplex.sql

  Summary:  How do you recover from any modification - when there are
			all sorts of other VALID modifications throughout the 
			database AND some time has passed.
  
  Date:     July 2010
		
  SQL Server Version: SQL Server 2008 SP1
------------------------------------------------------------------------------
  Written by Kimberly L. Tripp, SQLskills.com

  For more scripts and sample code, check out: 
    http://www.SQLskills.com

  This script is intended only as a supplement to demos and lectures
  written/presented by SQLskills.com  
  
  THIS CODE AND INFORMATION ARE PROVIDED "AS IS" WITHOUT WARRANTY OF 
  ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED 
  TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
  PARTICULAR PURPOSE.
============================================================================*/

------------------------------------------------------------------------------
-- ** STEP 1 Begin **
------------------------------------------------------------------------------
-- Create a new full backup to use as the base for restore operations:
USE master;
GO

BACKUP DATABASE SalesDB 
	TO DISK = N'C:\SQLskills\SalesDBBackup.bak'
	WITH NAME = N'Full Database Backup',
	DESCRIPTION = 'Starting point for recovery',
	INIT,
	STATS = 10;
GO

-- ** STEP 1 End **
------------------------------------------------------------------------------


------------------------------------------------------------------------------
-- ** STEP 2 Begin **
------------------------------------------------------------------------------
USE SalesDB;
GO

SELECT count(*) AS "Sales: Row Count", max(SalesID) AS "Sales: Max SalesID"
	FROM Sales 
	WITH (READUNCOMMITTED);
GO

SELECT getdate() AS "Date";
GO

-- Saved numbers:
--Sales: Row Count	Sales: Max SalesID
--6715221			6737977
-- Saved "Known good" Date/time: 2010-07-28 16:45:34.003

-- ** STEP 2 End **
------------------------------------------------------------------------------


------------------------------------------------------------------------------
-- ** STEP 2a Begin **
------------------------------------------------------------------------------
-- Step 2a
-- Perform - other modifications (ModifyData.sql)
-- This simulates other modifications that could be occurring in addition
-- to inserts...

-- ** STEP 2a End **
------------------------------------------------------------------------------


------------------------------------------------------------------------------
-- ** STEP 3 Begin **
------------------------------------------------------------------------------
-- Simulate a user dropping the unprotected table.
DROP TABLE Sales;
GO

-- ** STEP 3 End **
------------------------------------------------------------------------------


--- When did you know that things were bad??
SELECT getdate()
GO
-- Saved "Known BAD" Date/time: 2010-07-28 09:54:24.507


------------------------------------------------------------------------------
-- ** STEP 4 Begin **
------------------------------------------------------------------------------
-- Examine the database file structure.
EXEC sp_helpfile;
GO

-- ** STEP 4 End **
------------------------------------------------------------------------------


------------------------------------------------------------------------------
-- ** STEP 5 Begin **
------------------------------------------------------------------------------
-- Take the database offline for regular users. 
-- Put up a webpage... 
-- notify sales...
-- these all need to be considered and setup/created - PRIOR to the disaster!

USE master;
GO

ALTER DATABASE SalesDB
	SET RESTRICTED_USER
	WITH ROLLBACK IMMEDIATE;
GO

-- ** STEP 5 End **
------------------------------------------------------------------------------


------------------------------------------------------------------------------
-- ** STEP 6 Begin **
------------------------------------------------------------------------------
-- Backup the log (if database inaccessible use no_truncate)
BACKUP LOG SalesDB 
	TO DISK = N'C:\SQLskills\SalesDBBackup.bak'
	WITH NAME = N'Transaction Log Backup',
	DESCRIPTION = 'Getting everything to current point in time.', 
	STATS = 10;
GO

-- ** STEP 6 End **
------------------------------------------------------------------------------


------------------------------------------------------------------------------
-- ** STEP 7 Begin **
------------------------------------------------------------------------------
-- See what backups we have
RESTORE HEADERONLY FROM
	DISK = N'C:\SQLskills\SalesDBBackup.bak';
GO

-- Query msdb:
SELECT Backup_Start_Date, 
	[Name],
	[Description],
	First_LSN, 
	Last_LSN, 
	Backup_Finish_Date, 
	* -- OR Just *
FROM msdb.dbo.backupset AS s
    JOIN msdb.dbo.backupmediafamily AS m
        ON s.media_set_id = m.media_set_id
WHERE database_name = 'SalesDB'
ORDER BY 1 ASC

-- Use standard reports:
--	Standard Reports: Backup and Restore Events

-- Use 3rd party software?

-- ** STEP 7 End **
------------------------------------------------------------------------------


------------------------------------------------------------------------------
-- ** STEP 8 Begin **
------------------------------------------------------------------------------
USE master;
GO

RESTORE DATABASE SalesDB
	FROM DISK = N'C:\SQLskills\SalesDBBackup.bak'
	WITH NORECOVERY, STATS = 10, --REPLACE -- SQL doesn't allow you to restore over a database where the transaction log (the "tail" of the log) has not been backed up (unless you force it to with REPLACE)
	STOPAT = '2010-07-28 16:45:34.003' -- Do we know a point in time where things were good?
	
GO

RESTORE LOG SalesDB
	FROM DISK = N'C:\SQLskills\SalesDBBackup.bak'
	WITH FILE = 2, STATS = 10,
	STOPAT = '2010-07-28 16:45:34.003' -- Do we know a point in time where things were good?
	,NORECOVERY;
GO

RESTORE DATABASE SalesDB WITH RECOVERY, RESTRICTED_USER -- this means that YOU can get in but users cannot!
go
-- ** STEP 8 End **
------------------------------------------------------------------------------


------------------------------------------------------------------------------
-- ** STEP 9 Begin **
------------------------------------------------------------------------------
-- See how many rows existed at the restore time.
USE SalesDB;
GO

SELECT count(*) AS "Restored: Row Count", max(SalesID) AS "Restored: Max SalesID"
	FROM Sales 
	WITH (READUNCOMMITTED);
GO

-- Saved numbers:
--Restored: Row Count	Restored: Max SalesID
--6717365				6740121
-- ** STEP 9 End **
------------------------------------------------------------------------------



------------------------------------------------------------------------------
-- ** STEP 10 Begin **
------------------------------------------------------------------------------
-- Create a database snapshot that reflects the point in time to which
-- you recovered the database. This is the point to which you will compare
-- all tables in the recovered version.

USE [master]
GO

/****** Object:  Database [SalesDB]    Script Date: 06/17/2010 13:56:25 ******/
CREATE DATABASE [SalesDB_Snapshot] ON  PRIMARY 
( NAME = N'SalesDBData'
	, FILENAME = N'C:\Program Files\Microsoft SQL Server\MSSQL10.SQLDEV01\MSSQL\DATA\SalesDBData.ss')
AS SNAPSHOT OF SalesDB 
GO

-- ** STEP 10 End **
------------------------------------------------------------------------------



------------------------------------------------------------------------------
-- ** STEP 11 Begin **
------------------------------------------------------------------------------
-- Reseed the identity value, creating a gap so that we can let users back in.
USE SalesDB
go

DBCC CHECKIDENT ('Sales', RESEED, 6760000);
GO

USE master;
GO

-- Let users back into the database.
ALTER DATABASE SalesDB
	SET MULTI_USER;
GO

-- ** STEP 11 End **
------------------------------------------------------------------------------


------------------------------------------------------------------------------
-- ** STEP 12 Begin **
------------------------------------------------------------------------------
-- Examine the backups to see what the time interval is we need to consider.
SELECT [Name],
	Backup_Start_Date, 
	Backup_Finish_Date, 
	[Description],
	First_LSN, 
	Last_LSN, 
	* -- get all of the columns
FROM msdb.dbo.backupset AS s
    JOIN msdb.dbo.backupmediafamily AS m
        ON s.media_set_id = m.media_set_id
WHERE database_name = 'SalesDB'
ORDER BY 1 ASC;
GO
------------------------------------------------------------------------------
-- ** STEP 12 End **
------------------------------------------------------------------------------


------------------------------------------------------------------------------
-- ** STEP 13 Begin **
------------------------------------------------------------------------------
-- Create the second copy for investigative purposes
USE master;
GO

RESTORE DATABASE [SalesDB_Investigate] 
FROM DISK = N'C:\SQLskills\SalesDBBackup.bak' 
WITH FILE = 1,  
	MOVE N'SalesDBData' TO N'C:\Program Files\Microsoft SQL Server\MSSQL10.SQLDEV01\MSSQL\DATA\SalesDBData_Investigate.mdf',  
	MOVE N'SalesDBLog' TO N'C:\Program Files\Microsoft SQL Server\MSSQL10.SQLDEV01\MSSQL\DATA\SalesDBLog_Investigate.ldf',  
	STANDBY = N'C:\SQLskills\SalesDB_UNDO.bak',
	STOPAT = '2010-07-28 16:45:34.003',
	STATS, REPLACE;
GO

RESTORE LOG [SalesDB_Investigate] 
FROM DISK = N'C:\SQLskills\SalesDBBackup.bak' 
WITH FILE = 2,
	STANDBY = N'C:\SQLskills\SalesDB_UNDO.bak',
	STOPAT = '2010-07-28 16:45:34.003',
	STATS;
GO

-- The snapshot and our "investigative" copy should match:
-- See if we found any more records.
SELECT count(*) AS "Snapshot: Row Count", max(SalesID) AS "Snapshot: Max SalesID"
	FROM SalesDB_Snapshot.dbo.Sales 
	WITH (READUNCOMMITTED);
GO

SELECT count(*) AS "Investigate: Row Count", max(SalesID) AS "Investigate: Max SalesID"
	FROM SalesDB_investigate.dbo.Sales 
	WITH (READUNCOMMITTED);
GO

-- Saved numbers:
--Snapshot: Row Count	Snapshot: Max SalesID
--6717365	6740121

--Inv...Row Count	Snapshot: Max SalesID
--6717365	6740121

-- ** STEP 13 End **
------------------------------------------------------------------------------

-- Make some notes:
--Known Good: 2010-07-28 16:45:34.003

--Known Bad: 2010-07-28 16:47:31.243

------------------------------------------------------------------------------
-- ** STEP 14 Begin **
------------------------------------------------------------------------------
-- Automate the process of "inching" through the log and using tablediff...

USE master
go

sp_configure 'xp_cmdshell', 1
go
reconfigure
go

DECLARE @RestoredTime datetime = '2010-07-28 16:45:34.003'
DECLARE @CurrentTime datetime = @RestoredTime
DECLARE @CurrentTimeStr	nvarchar(25)
DECLARE @ExecutionRun int = 0
DECLARE @ExecutionRunStr nvarchar(3)
DECLARE @TableName sysname
DECLARE @ExecStr nvarchar(max)

DECLARE TablesToProcess CURSOR LOCAL STATIC 
FOR 
SELECT name 
FROM SalesDB_Investigate.sys.tables 
WHERE name NOT IN ('sysdiagrams') 

WHILE (SELECT object_id('SalesDB_Investigate.dbo.Sales')) IS NOT NULL
BEGIN
	SELECT @CurrentTime = dateadd(ms, 2000, @CurrentTime), @ExecutionRun = @ExecutionRun + 1
	SELECT @ExecutionRunStr = RIGHT('000' + convert(varchar, @ExecutionRun), 3)
	SELECT @CurrentTimeStr = convert(varchar(25), @CurrentTime, 126)
	RAISERROR('Processing a STOPAT time of %s', 10, 1, @CurrentTimeStr)
	SELECT @ExecStr = 
		N'USE Master; RESTORE LOG [SalesDB_Investigate]'
		+ N' FROM DISK = N''C:\SQLskills\SalesDBBackup.bak''' 
		+ N' WITH FILE = 2,'
		+ N'  STANDBY = N''C:\SQLskills\SalesDB_UNDO.bak'', '
		+ N'  STOPAT = ''' + convert(varchar(25), @CurrentTime, 126) + N''', STATS'
	--SELECT @ExecStr
	EXEC (@ExecStr)
	
	-- For EVERY Table, let's check the differences
	OPEN TablesToProcess 
	FETCH TablesToProcess INTO @TableName
	WHILE @@Fetch_status = 0
	BEGIN
		SELECT @ExecStr = 'xp_cmdshell ''"c:\program files\microsoft sql server\100\com\tablediff.exe" -sourceserver (local)\SQLDev01 -sourcedatabase SalesDB_Investigate -sourcetable ' + @TableName + ' -destinationserver (local)\SQLDev01 -destinationdatabase SalesDB_Snapshot -destinationtable ' + @TableName + ' -c -dt ' + @ExecutionRunStr + '_' + @TableName + '_Diffs -f C:\SalesDiffs\' + @ExecutionRunStr + '_' + @TableName + '_Diffs.sql'''
		--SELECT @ExecStr
		EXEC (@ExecStr)
		FETCH TablesToProcess INTO @TableName
	END	
	CLOSE TablesTOProcess
END
GO

-- ** STEP 14 End **
------------------------------------------------------------------------------


------------------------------------------------------------------------------
-- ** STEP 15 Begin **
------------------------------------------------------------------------------
-- Review the output scripts and use those to manipulate and migrate changes...
